import importlib.util

from click.exceptions import ClickException
from rich.console import Console

from .configs import configs


def ensure_env_file_exists(console: Console):
    if not configs.env_file.parent.exists() or not configs.env_file.exists():
        console.print(
            f"[bold red]Error: Directory '{configs.env_file.parent}' not initialized.[/bold red]"
        )
        raise ClickException("Please run the 'init' command first.")


def load_env_file_and_run_migration(migration_task_partial):
    try:
        module_spec = importlib.util.spec_from_file_location(
            configs.env_file.stem, configs.env_file
        )
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)

        module.run_migrations_online(migration_task_partial)

    except Exception:
        raise ClickException(f"Could not load 'env.py' at '{configs.env_file}'.")



import re
from graphlib import CycleError, TopologicalSorter
from pathlib import Path

from click.exceptions import ClickException
from rich.console import Console

from .configs import configs
from .schema import ActionEnum, MigrationScriptMetadata


def _get_migration_scripts() -> list[Path]:
    if not configs.versions_dir.exists():
        return []

    return [
        file
        for file in configs.versions_dir.iterdir()
        if file.is_file() and file.suffix == ".py"
    ]


def _parse_migration_script_headers(script_path: Path) -> MigrationScriptMetadata:
    with open(script_path) as file:
        content = file.read()

    revision_id_match = re.search(r"Revision ID:\s*(\w+)", content)
    revises_id_match = re.search(r"Revises:[^\S\n]*(\w*)", content)

    return MigrationScriptMetadata(
        path=script_path,
        revision_id=revision_id_match.group(1) if revision_id_match else None,
        revises_id=revises_id_match.group(1)
        if revises_id_match and revises_id_match.group(1)
        else None,
    )


def build_migration_run_plan(
    console: Console,
    target_revision: str,
    current_revision: str | None,
    verbose: bool = False,
) -> list[tuple[MigrationScriptMetadata, ActionEnum]]:
    migration_metadata = [
        _parse_migration_script_headers(script) for script in _get_migration_scripts()
    ]
    revision_id_to_metadata_mapping = {
        metadata.revision_id: metadata
        for metadata in migration_metadata
        if metadata.revision_id is not None
    }

    graph = {
        metadata.revision_id: {metadata.revises_id}
        if metadata.revises_id is not None
        else set()
        for metadata in migration_metadata
        if metadata.revision_id is not None
    }

    try:
        ts = TopologicalSorter(graph)
        full_ordered_path = list(ts.static_order())
    except CycleError as error:
        console.print("[bold red]Error: A circular dependency was detected![/bold red]")
        console.print(f"Cycle path: {' -> '.join(error.args[1])}")
        raise ClickException("Circular dependency detected in the migration history.")

    relative_match = re.match(r"^(HEAD)~(\d+)$", target_revision.upper())
    if relative_match:
        base_revision_name, number_of_steps_back = relative_match.groups()
        target_index = len(full_ordered_path) - int(number_of_steps_back)
        target_revision = (
            full_ordered_path[target_index - 1] if target_index > 0 else "base"
        )
        if verbose:
            console.print(
                f"Resolved '{relative_match.group(0)}' to target revision '{target_revision}'"
            )

    try:
        current_index = (
            full_ordered_path.index(current_revision) + 1
            if current_revision is not None
            else 0
        )
    except ValueError:
        raise ClickException(
            f"Current revision '{current_revision}' not found in migration history."
        )

    if target_revision == "head":
        target_index = len(full_ordered_path)
    elif target_revision == "base":
        target_index = 0
    else:
        try:
            target_index = full_ordered_path.index(target_revision) + 1
        except ValueError:
            raise ClickException(
                f"Target revision '{target_revision}' not found in migration history."
            )

    if target_index > current_index:
        migrations_to_run = [
            (revision_id_to_metadata_mapping[revision_id], ActionEnum.UPGRADE)
            for revision_id in full_ordered_path[current_index:target_index]
        ]
    elif target_index < current_index:
        migrations_to_run = [
            (revision_id_to_metadata_mapping[revision_id], ActionEnum.DOWNGRADE)
            for revision_id in reversed(full_ordered_path[target_index:current_index])
        ]
    else:
        migrations_to_run = []

    if verbose:
        console.print(
            f"[bold blue]Migration run plan created for '{current_revision}' -> '{target_revision}'[/bold blue]"
        )

    return migrations_to_run



import re
from enum import StrEnum
from pathlib import Path
from typing import Any

import click
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.traceback import install

try:
    import tomllib  # Python 3.11+
except ImportError:
    import tomli as tomllib  # type: ignore[import]

install(extra_lines=0, max_frames=10)


class VersionLevel(StrEnum):
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"
    ALPHA = "alpha"
    BETA = "beta"
    RC = "rc"
    DEV = "dev"
    RELEASE = "release"


class SupportedBackend(StrEnum):
    POETRY = "poetry"
    UV = "uv"


def load_config():
    pyproject_path = Path.cwd() / "pyproject.toml"
    if pyproject_path.exists():
        with open(pyproject_path, "rb") as f:
            pyproject_data = tomllib.load(f)
            config = pyproject_data.get("tool", {}).get("versioning", {}).get("files")

            backend = (
                pyproject_data.get("tool", {}).get("versioning", {}).get("backend")
            )
            if backend is None:
                raise RuntimeError(
                    "`pyproject.toml` file doesn't contain `tool.versioning.backend` property."
                )
            match SupportedBackend(backend):
                case SupportedBackend.POETRY:
                    version = (
                        pyproject_data.get("tool", {}).get("poetry", {}).get("version")
                    )
                case SupportedBackend.UV:
                    version = pyproject_data.get("project", {}).get("version")

            if version is None:
                raise RuntimeError(
                    "`pyproject.toml` file doesn't contain `version` property."
                )

            return version, config
    else:
        raise RuntimeError("Unable to locate `pyproject.toml` file.")


def parse_version(version) -> dict[str, str | Any]:
    VERSION_PATTERN = re.compile(
        r"^"
        r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"
        r"(?:(?P<pre_type>a|b|rc)(?P<pre_num>\d+))?"
        r"(?:\.post(?P<post>\d+))?"
        r"(?:\.dev(?P<dev>\d+))?"
        r"$"
    )

    pattern_match = VERSION_PATTERN.match(version)
    if pattern_match:
        return pattern_match.groupdict()
    else:
        raise ValueError(
            f"Invalid version format {version}. Expected format: MAJOR.MINOR.PATCH[preTYPE[preNUM][.dev]]"
        )


def find_substring_index(strings: list[str], substring: str) -> int:
    return next(i for i, string in enumerate(strings) if substring in string)


def find_variable_and_replace_value_in_file(
    file_name: str,
    variable_name: str,
    new_value: str,
    dry_run: bool,
    return_changed_line_number: bool = False,
) -> None | int:
    file = Path.cwd() / file_name
    text = file.read_text().splitlines(keepends=True)
    line_index = find_substring_index(text, variable_name)
    text[line_index] = text[line_index].replace(
        re.search(f"{variable_name}.*$", text[line_index]).group(0),
        variable_name + " = " + f'"{new_value}"',
    )

    if not dry_run:
        file.write_text("".join(text))

    if return_changed_line_number:
        return line_index + 1


def sync_version_in_different_files(
    version, config, dry_run: bool
) -> list[dict[str, int | str]]:
    files_synced = []
    if "version_variable" in config:
        for value in config.get("version_variable"):
            file_name = value.split(":")[0]
            variable_name = value.split(":")[1]

            line_number = find_variable_and_replace_value_in_file(
                file_name,
                variable_name,
                version,
                dry_run,
                return_changed_line_number=True,
            )

            files_synced.append(
                {
                    "file_path": file_name,
                    "line_number": line_number,
                }
            )

    return files_synced


def versioning(level: VersionLevel, current_version: str, dry_run: bool) -> str:
    version_parts = parse_version(current_version)

    match level:
        case VersionLevel.MAJOR:
            new_version = f"{int(version_parts['major']) + 1}.0.0"
        case VersionLevel.MINOR:
            new_version = (
                f"{version_parts['major']}.{int(version_parts['minor']) + 1}.0"
            )
        case VersionLevel.PATCH:
            new_version = f"{version_parts['major']}.{version_parts['minor']}.{int(version_parts['patch']) + 1}"
        case VersionLevel.ALPHA | VersionLevel.BETA | VersionLevel.RC:
            pre_type_map = {
                VersionLevel.ALPHA: "a",
                VersionLevel.BETA: "b",
                VersionLevel.RC: "rc",
            }

            if version_parts["pre_type"] == pre_type_map[level]:
                pre_num = int(version_parts["pre_num"] or 0) + 1
            else:
                pre_num = 1

            new_version = f"{version_parts['major']}.{version_parts['minor']}.{version_parts['patch']}{pre_type_map[level]}{pre_num}"
        case VersionLevel.DEV:
            if version_parts["pre_type"] and version_parts["pre_num"]:
                pre_parts = f"{version_parts['pre_type']}{version_parts['pre_num']}"
            else:
                pre_parts = ""

            dev_num = int(version_parts["dev"] or 0) + 1
            new_version = f"{version_parts['major']}.{version_parts['minor']}.{version_parts['patch']}{pre_parts}.dev{dev_num}"
        case VersionLevel.RELEASE:
            if (
                version_parts["pre_type"]
                or version_parts["pre_num"]
                or version_parts["dev"]
            ):
                new_version = f"{version_parts['major']}.{version_parts['minor']}.{version_parts['patch']}"
            else:
                raise ValueError(
                    f"The project is already on release version {current_version}"
                )

    if not dry_run:
        find_variable_and_replace_value_in_file(
            "pyproject.toml", "version", new_version, dry_run=False
        )

    return new_version


def display_update_summary(
    console: Console, files_synced: list[dict[str, int | str]]
) -> None:
    grid = Table.grid()
    grid.add_column(justify="center")

    for sync_infomation in files_synced:
        grid.add_row(
            f"Synced version updates in file [cyan]`{sync_infomation['file_path']}`[/cyan] at line [cyan]{sync_infomation['line_number']}[/cyan]"
        )

    console.print(
        Panel(grid, title="Version Update Summary", border_style="blue", padding=(1, 2))
    )


@click.command()
@click.argument(
    "level", type=click.Choice([vl.value for vl in VersionLevel], case_sensitive=False)
)
@click.option(
    "--show-summary",
    "-s",
    is_flag=True,
    help="Show what files changed",
)
@click.option(
    "--dry-run",
    "-d",
    is_flag=True,
    help="Show what would be done without making changes",
)
def main(level: str, show_summary: bool, dry_run: bool):
    console = Console()

    current_version, versioning_config = load_config()
    new_version = versioning(VersionLevel(level), current_version, dry_run)
    files_synced = sync_version_in_different_files(
        new_version, versioning_config, dry_run
    )

    if dry_run:
        console.print(
            f"[yellow]Dry Run[/yellow]: Bumped version: [bold cyan]{current_version}[/bold cyan] :arrow_right: [bold green]{new_version}[/bold green]"
        )
    else:
        console.print(
            f"Bumped version: [bold cyan]{current_version}[/bold cyan] :arrow_right: [bold green]{new_version}[/bold green]"
        )

    if show_summary:
        display_update_summary(console, files_synced)


if __name__ == "__main__":
    main()



import importlib.util
import re
from datetime import UTC, datetime

from click.exceptions import ClickException
from mako.template import Template
from qdrant_client import QdrantClient
from rich.console import Console
from rich.table import Table

from .configs import configs
from .graph import build_migration_run_plan
from .schema import ActionEnum
from .templates import TEMPLATE_ENV_PY, TEMPLATE_README, TEMPLATE_SCRIPT_MAKO
from .utils import ensure_env_file_exists


def get_current_revision() -> str | None:
    if not configs.version_index_file.exists():
        return None

    with open(configs.version_index_file) as index_file:
        return index_file.read().strip() or None


def _set_current_revision(revision_id: str | None):
    configs.version_index_file.parent.mkdir(exist_ok=True, parents=True)
    with open(configs.version_index_file, "w") as index_file:
        index_file.write(revision_id or "")


def run_migration_task(
    console: Console,
    client: QdrantClient,
    target_revision: str,
    dry_run: bool = False,
    verbose: bool = False,
):
    current_revision = get_current_revision()

    migration_run_plan = build_migration_run_plan(
        console=console,
        target_revision=target_revision,
        current_revision=current_revision,
        verbose=verbose,
    )

    if len(migration_run_plan) == 0:
        console.print("[bold green]Database is already up-to-date.[/bold green]")
        return

    console.print(
        f"Migrating from '{current_revision if current_revision is not None else 'base'}' to '{target_revision}'"
    )

    table = Table()
    table.add_column("Action", style="cyan", no_wrap=True)
    table.add_column("Revision ID", style="magenta")
    for migration, action in migration_run_plan:
        match action:
            case ActionEnum.UPGRADE:
                action_style = "bold green"
            case ActionEnum.DOWNGRADE:
                action_style = "bold red"

        table.add_row(
            f"[{action_style}]{action.upper()}[/{action_style}]", migration.revision_id
        )

    console.print(table)

    if dry_run:
        console.print(
            "\n[bold yellow]Dry run successful. No changes were made.[/bold yellow]"
        )
        return

    console.print("\n[bold]Applying migrations...[/bold]")
    for migration, action in migration_run_plan:
        try:
            module_spec = importlib.util.spec_from_file_location(
                migration.path.stem, migration.path
            )
            module = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module)

            action_function = getattr(module, action)

            with console.status(
                f"[bold cyan]Running {action.upper()} for {migration.revision_id}...[/bold cyan]"
            ):
                action_function(client)

            match action:
                case ActionEnum.UPGRADE:
                    _set_current_revision(migration.revision_id)
                case ActionEnum.DOWNGRADE:
                    _set_current_revision(migration.revises_id)

            if verbose:
                console.print(f"[gray50]Done: {migration.revision_id}[/gray50]")

        except (ImportError, AttributeError) as error:
            raise ClickException(
                f"Invalid migration script '{migration.path.name}': {error}"
            )
        except Exception:
            if verbose:
                console.print_exception(show_locals=True)

            raise ClickException(
                f"[bold red]An error occurred during execution of '{migration.revision_id}'. Aborting migration.[/bold red]"
            )

    console.print("[bold green]Migration finished successfully.[/bold green]")


def create_new_migration(
    console: Console,
    short_description: str,
    long_description: None | str,
    verbose: bool = False,
):
    ensure_env_file_exists(console=console)

    current_head = get_current_revision()
    if current_head is None:
        path = build_migration_run_plan(
            console=console,
            target_revision="head",
            current_revision=None,
            verbose=verbose,
        )
        current_head = path[-1][0].revision_id if path else None

    revises_id = current_head or ""

    new_revision_id = datetime.now(UTC).strftime("%Y%m%d%H%M%S")
    sanitized_short_description = (
        re.sub(r"[\s\W]+", "_", short_description).lower().strip("_")
    )

    migration_file_path = (
        configs.versions_dir
        / f"{new_revision_id}_{sanitized_short_description[:39].strip('_')}.py"
    )

    try:
        if not configs.template_file.exists():
            raise ClickException(
                f"Template file not found at '{configs.template_file}'"
            )

        template = Template(filename=str(configs.template_file))  # noqa: S702
        rendered_migration_script = template.render(
            revision=new_revision_id,
            revises=revises_id,
            created_at=datetime.now(UTC).isoformat(),
            description=short_description
            + ("\n\n" + long_description if long_description else ""),
        )
        with open(migration_file_path, "w") as file:
            file.write(rendered_migration_script)

        console.print("[bold green]Created new migration file.[/bold green]")

    except Exception as error:
        raise ClickException(f"Failed to create new migration file. {error}")


def initialize_orthant_env(console: Console, verbose: bool = False):
    if not configs.versions_dir.exists():
        configs.versions_dir.mkdir(parents=True, exist_ok=True)

        if verbose:
            console.print(f"[green]Created '{configs.versions_dir}'.[/green]")
    else:
        if verbose:
            console.print(f"[yellow]'{configs.versions_dir}' already exists.[/yellow]")

    if not configs.env_file.exists():
        with open(configs.env_file, "w") as env_file:
            env_file.write(TEMPLATE_ENV_PY)

        if verbose:
            console.print(f"[green]Created '{configs.env_file}'.[/green]")
    else:
        if verbose:
            console.print(f"[yellow]'{configs.env_file}' already exists.[/yellow]")

    if not configs.template_file.exists():
        with open(configs.template_file, "w") as template_file:
            template_file.write(TEMPLATE_SCRIPT_MAKO)

        if verbose:
            console.print(f"[green]Created '{configs.template_file}'.[/green]")
    else:
        if verbose:
            console.print(f"[yellow]'{configs.template_file}' already exists.[/yellow]")

    if not configs.readme_file.exists():
        with open(configs.readme_file, "w") as readme_file:
            readme_file.write(TEMPLATE_README)

        if verbose:
            console.print(f"[green]Created '{configs.readme_file}'.[/green]")
    else:
        if verbose:
            console.print(f"[yellow]'{configs.readme_file}' already exists.[/yellow]")

    console.print(
        f"[bold green]Successfully initialized Orthant at '{configs.version_index_file.parent}'.[/bold green]"
    )
    console.print(
        "Next steps: Edit [bold]env.py[/bold] to configure your Qdrant connection."
    )



from functools import partial

import click
from rich.console import Console

from . import __version__
from .commands import (
    create_new_migration,
    get_current_revision,
    initialize_orthant_env,
    run_migration_task,
)
from .utils import ensure_env_file_exists, load_env_file_and_run_migration


@click.group()
def main() -> None:
    pass


@main.command()
def version() -> None:
    console = Console()
    console.print(f"Orthant CLI version: [green]{__version__}[/green]")


@main.command()
@click.option(
    "-m",
    "--message",
    required=True,
    multiple=True,
    help="Migration message. The first is the subject, subsequent are the body.",
)
@click.option("--verbose", is_flag=True, help="Enable verbose mode to echo steps.")
def new(message: tuple[str, ...], verbose: bool) -> None:
    console = Console()

    short_description = message[0]
    long_description = "\n".join(message[1:]) if len(message) > 1 else ""

    create_new_migration(
        console=console,
        short_description=short_description,
        long_description=long_description,
        verbose=verbose,
    )


@main.command()
@click.option("--verbose", is_flag=True, help="Enable verbose mode to echo steps.")
def init(verbose: bool) -> None:
    console = Console()

    initialize_orthant_env(console=console, verbose=verbose)


@main.command()
def current() -> None:
    console = Console()

    ensure_env_file_exists(console=console)

    current_revision = get_current_revision()
    if current_revision is not None:
        console.print(f"Current Revision: [bold cyan]{current_revision}[/bold cyan]")
    else:
        console.print(
            "Database is at [bold cyan]'base'[/bold cyan] (no migrations applied)."
        )


@main.command()
@click.argument("revision", nargs=1, default="head", type=click.STRING)
@click.option("--dry-run", is_flag=True, help="Simulate the migration.")
@click.option("--verbose", is_flag=True, help="Enable verbose mode to echo steps.")
def upgrade(revision: str, dry_run: bool, verbose: bool) -> None:
    console = Console()

    ensure_env_file_exists(console=console)

    task = partial(
        run_migration_task,
        console=console,
        target_revision=revision,
        dry_run=dry_run,
        verbose=verbose,
    )
    load_env_file_and_run_migration(task)


@main.command()
@click.argument("revision", nargs=1, default="HEAD~1", type=click.STRING)
@click.option("--dry-run", is_flag=True, help="Simulate the migration.")
@click.option("--verbose", is_flag=True, help="Enable verbose mode to echo steps.")
def downgrade(revision: str, dry_run: bool, verbose: bool) -> None:
    console = Console()

    ensure_env_file_exists(console=console)

    task = partial(
        run_migration_task,
        console=console,
        target_revision=revision,
        dry_run=dry_run,
        verbose=verbose,
    )
    load_env_file_and_run_migration(task)



from enum import StrEnum
from pathlib import Path

from pydantic import BaseModel


class ActionEnum(StrEnum):
    UPGRADE = "upgrade"
    DOWNGRADE = "downgrade"


class MigrationScriptMetadata(BaseModel):
    path: Path
    revision_id: str | None
    revises_id: str | None



import tomllib
from pathlib import Path

from pydantic import BaseModel


class Config(BaseModel):
    versions_dir: Path
    version_index_file: Path
    template_file: Path
    env_file: Path
    readme_file: Path


def load_configs() -> Config:
    orthant_config_file = Path("orthant.toml")
    pyproject_file = Path("pyproject.toml")

    if not orthant_config_file.exists():
        if not pyproject_file.exists():
            raise FileNotFoundError("Unable to locate configuration file.")
        else:
            config_file = pyproject_file
    else:
        config_file = orthant_config_file

    with open(config_file, "rb") as file:
        raw_orthant_config: dict = tomllib.load(file).get("tool", {}).get("orthant", {})

        raw_orthant_base_dir: str = raw_orthant_config.get("base_dir", "orthant")
        parsed_orthant_base_dir: Path = Path(raw_orthant_base_dir)

        parsed_orthant_versions_dir: Path = parsed_orthant_base_dir / "versions"
        parsed_orthant_version_index_file: Path = (
            parsed_orthant_base_dir / "VERSION_INDEX"
        )
        parsed_orthant_template_file: Path = parsed_orthant_base_dir / "script.py.mako"
        parsed_orthant_env_file: Path = parsed_orthant_base_dir / "env.py"
        parsed_orthant_readme_file: Path = parsed_orthant_base_dir / "README.md"

        return Config(
            versions_dir=parsed_orthant_versions_dir,
            version_index_file=parsed_orthant_version_index_file,
            template_file=parsed_orthant_template_file,
            env_file=parsed_orthant_env_file,
            readme_file=parsed_orthant_readme_file,
        )


configs = load_configs()



TEMPLATE_ENV_PY = """import os

from qdrant_client import QdrantClient

# --- Configuration ---
# This is where you connect the migration tool to your application's settings.
# You might import your standard Pydantic settings object here.
# e.g.: from app.settings import settings

QDRANT_URL = os.getenv("QDRANT_URL", "http://localhost:6333")
QDRANT_API_KEY = os.getenv("QDRANT_API_KEY")


def get_qdrant_client() -> QdrantClient:
    \"\"\"Creates and returns an instance of the asynchronous Qdrant client.

    Customize this to add timeouts, custom headers, or use gRPC if needed.
    \"\"\"
    return QdrantClient(url=QDRANT_URL, api_key=QDRANT_API_KEY)


def run_migrations_online(run_migration_fn) -> None:
    \"\"\"The main execution context for migrations.

    It sets up the client, runs the migration task, and ensures cleanup.

    Args:
        run_migration_fn (_type_): The callback provided by the migration runner.
    \"\"\"
    client = get_qdrant_client()
    try:
        # Execute the migration task with the configured client
        run_migration_fn(client=client)
    finally:
        # Ensure connections are always closed, even on error
        client.close()
"""


TEMPLATE_SCRIPT_MAKO = """\"\"\"${description}

Revision ID: ${revision}
Revises: ${revises}
Create Date: ${created_at}
\"\"\"

from qdrant_client import QdrantClient, models

# revision identifiers, used by the migration tool
revision = "${revision}"
down_revision = "${revises or None}"


def upgrade(client: QdrantClient) -> None:
    # ### commands auto-generated by Orthant - please adjust! ###
    pass
    # ### end Orthant commands ###


def downgrade(client: QdrantClient) -> None:
    # ### commands auto-generated by Orthant - please adjust! ###
    pass
    # ### end Orthant commands ###
"""

TEMPLATE_README = """# Orthant - Qdrant Migration Tool

This directory contains migration scripts for ensuring the Qdrant vector database schema (collections, indexes, payload schema) matches the application code.

## Structure

* `versions/`: Holds the individual migration scripts.
* `env.py`: Configuration file. This is where you customize how the tool connects to Qdrant.
* `script.py.mako`: The template used when generating new migrations.
* `.VERSION_INDEX`: Tracks the currently applied revision ID. DO NOT EDIT MANUALLY.
"""
